from sqlalchemy import Column, Boolean, Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import JSON, JSONB, UUID
from sqlalchemy.ext.declarative import declarative_base

from spintop.models import PersistenceIDRecord, SerializedPersistenceRecord
from spintop.models.serialization import TYPE_KEY

Base = declarative_base()

fields = PersistenceIDRecord.fields()

# Native table fields. The rest is in the 'index' JSONB column of the table

# Those are replicated in the serialized version as-is
writeable_top_level_fields = { fields.uuid, fields.is_deleted }

# Those are simply indexed when the index data changes.
# Datetime is omitted since it would require serialization/deserialization from postgresql TIMESTAMP
top_level_fields = writeable_top_level_fields | { fields.start_datetime }

class DataBlob(Base):
    __tablename__ = 'data_blobs'
    blob_id = Column(UUID(as_uuid=True), primary_key=True, unique=True, nullable=False)
    _group = Column(String, primary_key=True, nullable=True)
    data = Column(JSON)

class RecordIndex(Base):
    __tablename__ = 'records_index'
    # top level fields
    uuid = Column(String, primary_key=True)
    is_deleted = Column(Boolean)
    start_datetime = Column(DateTime(timezone=True))
    #

    _group = Column(String, primary_key=True, nullable=True)
    _type = Column(String)

    index = Column(JSONB)
    data_blob_id = Column(UUID(as_uuid=True), ForeignKey(DataBlob.blob_id, ondelete='SET NULL'), nullable=True)

    data_blob = relationship(DataBlob)

    @classmethod
    def create(cls, record, _group=None, data_blob_id=None):
        return cls(
            index=record.index, 
            _type=record.type_,
            _group=_group,
            data_blob_id=data_blob_id,
            **{field.name_: field.get_value(record.index) for field in top_level_fields}
        )

    def to_serialized_record(self, with_data=True):
        content = {
            TYPE_KEY: self._type,
            'index': self.index
        }

        for field in writeable_top_level_fields:
            # Source of truth are the top level fields.
            # Replicate them inside the index, but serialize them first
            field.set_value(content['index'], getattr(self, field.name_))

        if with_data:
            content['data'] = self.data_blob.data
        return SerializedPersistenceRecord(content)



import zlib
import json

class MultiEncoder():
    def __init__(self, parts):
        self._parts = parts

    def encode(self, obj):
        result = obj
        for part in self._parts:
            result = part.encode(result)
        return result

    def decode(self, result):
        obj = result
        for part in reversed(self._parts):
            obj = part.decode(obj)
        return obj

class JSONEncoder():
    def __init__(self, encoder_cls=None, sort_keys=True, encoding='utf-8'):
        self._encoder_kwargs = dict(
            cls=encoder_cls,
            sort_keys=sort_keys
        )
        self.bytes_encoding = encoding

    def encode(self, obj):
        return json.dumps(obj, **self._encoder_kwargs).encode(self.bytes_encoding)
    
    def decode(self, result):
        return json.loads(result.decode(self.bytes_encoding))

class ZlibCompressor():
    def __init__(self, level=-1):
        self.level = level

    def encode(self, obj):
        return zlib.compress(obj, level=self.level)

    def decode(self, result):
        return zlib.decompress(result)

def compressed_json_encoder_factory(encoder_cls=None, sort_keys=True, encoding='utf-8', compression_level=-1):
    return MultiEncoder([
        JSONEncoder(encoder_cls=encoder_cls, sort_keys=sort_keys, encoding=encoding),
        ZlibCompressor(level=compression_level)
    ])



from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any, Union

from spintop.models import PersistenceRecord, PersistenceIDRecord, BaseDataClass

class SpintopEnvMessage(BaseDataClass):
    env: Optional[Dict[str, Optional[Any]]]

    def create_message_env(self, local_env):
        """Create a copy of the local env with the attributes contained in this message"""
        env = self.env
        if env is None:
            env = {}
        return local_env.copy(**env)

    @classmethod
    def from_message(cls, other_env_message, **kwargs):
        obj = cls(env=other_env_message.env, **kwargs)
        return obj

class PersistenceRecordMessage(SpintopEnvMessage):
    record: PersistenceRecord

class PersistenceRecordsUpdateMessage(SpintopEnvMessage):
    updated_ids: List[PersistenceIDRecord]
    update_all: bool = False

    def merge(self, other):
        return self.__class__(
            env=self.env,
            update_all=self.update_all or other.update_all,
            updated_ids=self.updated_ids + other.updated_ids
        )

    @classmethod
    def create(cls, **kwargs):
        defaults = dict(
            updated_ids = [],
            update_all = False,
            env = None
        )
        defaults.update(kwargs)
        return cls(**defaults)



import logging
import warnings

LEVEL = logging.INFO
VERBOSE = (LEVEL == logging.DEBUG)
CORE_LOGGER = logging.getLogger('sptp')
FORMAT = '%(relativeCreated)-10d[ms] %(name)-30s:%(levelname)-5s: %(message)s'

STREAM_HANDLER = logging.StreamHandler()
STREAM_HANDLER.setFormatter(logging.Formatter(FORMAT))

def _logger(name):
    return CORE_LOGGER.getChild(name)

def setup_logging():
    root = logging.getLogger()
    if root.handlers:
        for handler in root.handlers:
            root.removeHandler(handler)

    logging.basicConfig(format=FORMAT, level=logging.WARNING)

    CORE_LOGGER.setLevel(LEVEL)
    
    # if STREAM_HANDLER not in CORE_LOGGER.handlers:
    #     CORE_LOGGER.addHandler(STREAM_HANDLER)

def deprecated(reason):
    warnings.warn(reason, DeprecationWarning)



from itertools import chain

from collections import OrderedDict, defaultdict
from collections.abc import Mapping
from typing import NamedTuple, List, Dict, Any, Optional

from marshmallow import Schema, post_dump, pre_load

from spintop.logs import deprecated
from spintop.utils import GetRecursive, GetRecursiveMixin, dict_ops

from ..serialization import (
    is_serialized_type_of,
    type_of
)

from ..view import DataClassPrimitiveView, ComplexPrimitiveView
from ..meta import FieldsOf

from ..persistence import PersistenceRecord, PersistenceRecordCollection

from ._base import (
    model_property,
    BaseDataClass, 
    TestIDRecord, 
    TestRecordSummary, 
    FeatureRecord, 
    DefaultPrimitiveView, 
    compute_stats
)

SUPPORT_MISSING_TOP_LEVEL_RECORD = False

def normalize_features_len(dict_array):
    new_dict = {}
    max_column_key_len = max([len(key) for key in dict_array])
    for key, value in dict_array.items():
        if len(key) < max_column_key_len:
            key = key + ('',)*(max_column_key_len - len(key))
        new_dict[key] = value
    return new_dict

class SpintopTestRecordData(BaseDataClass):
    summary: TestRecordSummary = None
    features: List[FeatureRecord] = list


class SpintopTestRecord(PersistenceRecord):

     # PersistenceRecord interface
    index: TestIDRecord = None
    data: SpintopTestRecordData = SpintopTestRecordData

    ## Alias properties ##

    @property
    def summary(self) -> TestRecordSummary:
        return self.data.summary

    @summary.setter
    def summary(self, value):
        self.test_id = value.test_id
        self.data.summary = value
        self._broadcast_test_id(self.features)

    @property
    def test_id(self):
        return self.index

    @test_id.setter
    def test_id(self, value):
        self.index = value

    @property
    def features(self):
        return self.data.features if self.data.features else tuple()
    
    @features.setter
    def features(self, features):
        self._broadcast_test_id(features)
        self.data.features = tuple(sorted(features, key=lambda feature: feature.index))
        

    ## Other properties ##

    @property
    def all_features(self):
        return (self.summary,) + tuple(self.features)

    def compute_stats(self):
        if not self.summary.feature_count:
            compute_stats(self.all_features)

    @property
    def test_record(self):
        deprecated('Please use the `summary` attribute instead.')
        return self.summary

    @property
    def test_uuid(self):
        return self.test_id.uuid
    
    def __hash__(self):
        return hash(self.test_uuid)

    def __eq__(self, other):
        return self.summary == other.summary and self.features == other.features
    
    def __repr__(self):
        return '{}(summary={!r}, features=[...]*{})'.format(self.__class__.__name__, self.summary, len(self.features))

    def remove_duplicate_features(self):
        feature_names = set()
        features = []
        for feature in self.features:
            if feature.name in feature_names:
                continue
            else:
                features.append(feature)
                feature_names.add(feature.name)
        
        self.features = features
    
    def _broadcast_test_id(self, features):
        """Uses the top level test_id and sets it for all features, ensuring
        that all features have the same TestIDRecord reference."""
        for feature in features:
            feature.test_id = self.test_id
    
    def reindex(self):
        for index, feature in enumerate(self.features):
            feature.index = index

    def normalize_outcomes(self):
        parents = tuple()
        last_feature = None
        for feature in self.all_features:
            while feature.depth > len(parents):
                if last_feature.name not in feature.ancestors:
                    raise Exception('Unable to normalize outcomes: feature order does not match ancestors. {!r} should be in ancestors: {!r}'.format(
                        last_feature.name,
                        feature.ancestors
                    )) 
                parents = parents + (last_feature,)

            if feature.depth < len(parents):
                parents = parents[:feature.depth]

            outcome = feature.outcome

            for parent in reversed(parents):
                outcome.impose_upon(parent.outcome)
                outcome = parent.outcome

            last_feature = feature

            
    def fill_missing_from_source(self, fill_source, on_fill=None):
        max_current_feature_index = max(self.features, key=lambda f: f.index)
        new_features_len = len(fill_source)
        
        assert new_features_len >= max_current_feature_index.index, "Fill source must be same length or bigger than this test_record max index."
        
        new_features = [None]*new_features_len
        
        if on_fill is None:
            on_fill = lambda obj: obj.copy()
            
        def fill_range(start, end):
            for i in range(start, end):
                new_features[i] = on_fill(fill_source[i])
        
        current_index = 0
        for feature in sorted(self.features, key=lambda f: f.index):
            fill_range(current_index, feature.index)
            new_features[feature.index] = feature # Keep this feature
            current_index = feature.index + 1
        
        fill_range(current_index, new_features_len) # Fill end
        
        self.features = new_features
                
    def find_feature(self, condition, start_index=0):
        try:
            return next(feat for feat in self.all_features[start_index:] if condition(feat))
        except StopIteration:
            raise ValueError('No feature matched condition.')
    
    def add_tag(self, key, value=True):
        self.test_id.add_tag(key, value=value)
        
    def remove_tag(self, key):
        self.test_id.remove_tag(key)

spintop_test_record_features = FieldsOf(SpintopTestRecord).data.features
spintop_test_record_summary = FieldsOf(SpintopTestRecord).data.summary
spintop_test_record_test_id = FieldsOf(SpintopTestRecord).index

class SpintopTestRecordCollection(PersistenceRecordCollection):
    records: List[SpintopTestRecord] = list

    @property
    def test_uuids(self):
        return [record.uuid for record in self.records]
            
    def are_records_unique(self):
        test_uuids = self.test_uuids
        return len(test_uuids) == len(set(test_uuids))
        
    def avg_feature_count(self):
        feature_counts = [len(record.features) for record in self.iter_records()]
        return sum(feature_counts)/len(feature_counts)

class SpintopTestRecordView(object):
    feature_prefix = ('features', lambda feat: type_of(feat), lambda feat: feat.complete_name)
    data_prefix = () 
    default_view = DefaultPrimitiveView()
    
    def __init__(self, feature_mapping=None, include_test_id=True):
        self.include_test_id = include_test_id

        if feature_mapping:
            self.feature_view = DataClassPrimitiveView(feature_mapping)
        else:
            self.feature_view = self.default_view

    def apply(self, record, flatten_dict=True):
        data = OrderedDict()
        
        if self.include_test_id:
            test_id_data = self.apply_default_test_id(record, flatten_dict=flatten_dict, attribute_error_policy='raise')
            dict_ops.update(data, test_id_data)
        
        for feature in record.features:
            feature_data = self.feature_view.apply(feature, flatten_dict=flatten_dict)
            dict_ops.update(data, feature_data)
        
        data = dict_ops.replace_defaults(data)

        return data
    
    def apply_default_feature(self, feature, key_prefix=None, **apply_kwargs):
        if key_prefix is None: key_prefix = self.feature_prefix
        return self.feature_view.apply(feature, key_prefix=key_prefix, **apply_kwargs)
    
    def apply_default_test_id(self, record, key_prefix=None, **apply_kwargs):
        if key_prefix is None: key_prefix = self.data_prefix
        data = self.default_view.apply(record.summary, key_prefix=key_prefix, **apply_kwargs)
        return data

class SpintopTestRecordBuilder(NamedTuple):
    summary: TestRecordSummary
    features: List[FeatureRecord]
    
    def build(self):
        record = SpintopTestRecord(summary=self.summary, features=sorted(self.features, key=lambda feature: feature.index))
        record.compute_stats()
        return record

class SpintopSerializedTestRecord(object):
    def __init__(self, serialized_record=None):
        if not serialized_record: serialized_record = self.create_empty_serialized_record()
        self._record = serialized_record

        if not self.features:
            self.features = []

    def create_empty_serialized_record(self):
        record = SpintopTestRecord()
        schema = SpintopTestRecord.get_schema()
        return schema().dump(record)

    @property
    def summary(self):
        return spintop_test_record_summary.get_value(self._record)

    @summary.setter
    def summary(self, value):
        spintop_test_record_summary.set_value(self._record, value)

    @property
    def test_record(self):
        deprecated('Please use the `summary` attribute instead.')
        return self.summary

    @property
    def features(self):
        return spintop_test_record_features.get_value(self._record)

    @features.setter
    def features(self, value):
        spintop_test_record_features.set_value(self._record, value)

    @property
    def test_id(self):
        return spintop_test_record_test_id.get_value(self._record)

    @property
    def test_uuid(self):
        return self.test_id['uuid']

    @property
    def all_features(self):
        return (self.summary,) + tuple(self.features)

    def as_dict(self):
        return self._record

    def deserialize(self, serializer):
        builder = SpintopTestRecordBuilder(
            summary=serializer.deserialize(self.summary), 
            features=tuple(serializer.deserialize(feat) for feat in self.features)
        )
        return builder.build()

    def __repr__(self):
        return '{}(summary={!r}, features=[...]*{})'.format(self.__class__.__name__, self.summary, len(self.features))

    def __eq__(self, other):
        return self._record == other._record

class SpintopSerializedTestRecordCollection(object):
    records: List[SpintopSerializedTestRecord]

    def __init__(self, records:SpintopSerializedTestRecord = None):
        self.records = [
            SpintopSerializedTestRecord(record) if not isinstance(record, SpintopSerializedTestRecord) else record 
            for record in records
        ]

    @classmethod
    def from_features(cls, features):
        return cls(records=serialized_features_to_record_list(features))

    def deserialize(self, serializer):
        return (record.deserialize(serializer) for record in self.records)

TestUUID = FieldsOf(FeatureRecord).test_id.uuid
def serialized_features_to_record_list(features):
    test_uuid_record_map = OrderedDict()
    
    for feature in features:
        found_summary = None
        found_feature = None

        if is_serialized_type_of(feature, TestRecordSummary):
            found_summary = feature
        else:
            found_feature = feature
        
        test_uuid = TestUUID.get_value(feature)

        record = test_uuid_record_map.get(test_uuid, None)
        if record is None:
            record = SpintopSerializedTestRecord()
        
        if found_summary:
            if record.summary is not None:
                raise ValueError('Two features match as a top level record for the same record: {!r} and {!r}'.format(record.test_record, found_summary))
            record.summary = found_summary
        else:
            record.features.append(found_feature)

        test_uuid_record_map[test_uuid] = record
    
    records = list(test_uuid_record_map.values())

    for record in records:
        if record.summary is None:
            # Find a valid test_id from features.
            if SUPPORT_MISSING_TOP_LEVEL_RECORD:
                test_id = None
                if record.features:
                    test_id = record.features[0]['test_id']
                else:
                    raise ValueError('Empty Test Record.')
                record.summary = TestRecordSummary.null(test_id=test_id, original=False).as_dict()
            else:
                raise RuntimeError('Missing top level record from SpintopTestRecord {!r}'.format(record))

    return records

        
        
    


from collections.abc import Mapping
from marshmallow import Schema, ValidationError


class OneOfSchema(Schema):
    """
    This is a special kind of schema that actually multiplexes other schemas
    based on object type. When serializing values, it uses get_obj_type() method
    to get object type name. Then it uses `type_schemas` name-to-Schema mapping
    to get schema for that particular object type, serializes object using that
    schema and adds an extra "type" field with name of object type.
    Deserialization is reverse.
    Example:
        class Foo(object):
            def __init__(self, foo):
                self.foo = foo
        class Bar(object):
            def __init__(self, bar):
                self.bar = bar
        class FooSchema(marshmallow.Schema):
            foo = marshmallow.fields.String(required=True)
            @marshmallow.post_load
            def make_foo(self, data, **kwargs):
                return Foo(**data)
        class BarSchema(marshmallow.Schema):
            bar = marshmallow.fields.Integer(required=True)
            @marshmallow.post_load
            def make_bar(self, data, **kwargs):
                return Bar(**data)
        class MyUberSchema(marshmallow.OneOfSchema):
            type_schemas = {
                'foo': FooSchema,
                'bar': BarSchema,
            }
            def get_obj_type(self, obj):
                if isinstance(obj, Foo):
                    return 'foo'
                elif isinstance(obj, Bar):
                    return 'bar'
                else:
                    raise Exception('Unknown object type: %s' % repr(obj))
        MyUberSchema().dump([Foo(foo='hello'), Bar(bar=123)], many=True)
        # => [{'type': 'foo', 'foo': 'hello'}, {'type': 'bar', 'bar': 123}]
    You can control type field name added to serialized object representation by
    setting `type_field` class property.
    """

    type_field = "type"
    type_field_remove = True
    type_schemas = []

    def get_obj_type(self, obj):
        """Returns name of object schema"""
        return obj.__class__.__name__

    def dump(self, obj, *, many=None, **kwargs):
        errors = {}
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if not many:
            try:
                result = result_data = self._dump(obj, **kwargs)
            except ValidationError as error:
                result_errors = error.normalized_messages()
                result_data.append(error.valid_data)
        else:
            for idx, o in enumerate(obj):
                try:
                    result = self._dump(o, **kwargs)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.normalized_messages()
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=obj, valid_data=result)
            raise exc

    def _dump(self, obj, *, update_fields=True, **kwargs):

        # Returns None only if the field is present, but the class is not registered.
        obj_type = self.get_obj_type(obj)
        
        if isinstance(obj, Mapping):
            # return mapping as-is, the obj_type is defined, but possibly not understood.
            return obj
        
        if not obj_type:
            raise ValidationError(
                {"_schema": "Unknown object class: %s" % obj.__class__.__name__},
            )

        type_schema = self.type_schemas.get(obj_type)

        if not type_schema:
            raise ValidationError({"_schema": "Unsupported object type: %s" % obj_type})

        schema = type_schema if isinstance(type_schema, Schema) else type_schema()

        schema.context.update(getattr(self, "context", {}))

        if isinstance(schema, OneOfSchema):
            if isinstance(obj, Mapping):
                # Already dumped and as an obj_type.
                result = obj
            else:
                result = super(OneOfSchema, schema).dump(obj, many=False, **kwargs)
        else:
            result = schema.dump(obj, many=False, **kwargs)

        if result is not None:
            result[self.type_field] = obj_type
        return result

    def load(self, data, *, many=None, partial=None, unknown=None):
        errors = {}
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if partial is None:
            partial = self.partial
        if not many:
            try:
                result = result_data = self._load(
                    data, partial=partial, unknown=unknown
                )
                #  result_data.append(result)
            except ValidationError as error:
                result_errors = error.normalized_messages()
                result_data.append(error.valid_data)
        else:
            for idx, item in enumerate(data):
                try:
                    result = self._load(item, partial=partial)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.normalized_messages()
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=data, valid_data=result)
            raise exc

    def _load(self, data, *, partial=None, unknown=None):
        if not isinstance(data, dict):
            raise ValidationError({"_schema": "Invalid data type: %s" % data})

        data = dict(data)
        unknown = unknown or self.unknown

        context = getattr(self, "context", {})

        data_type = data.get(self.type_field)

        if not data_type:
            raise ValidationError(
                {self.type_field: ["Missing data for required field."]}
            )

        try:
            type_schema = self.type_schemas.get(data_type)
        except TypeError:
            # data_type could be unhashable
            raise ValidationError({self.type_field: ["Invalid value: %s" % data_type]})

        if not type_schema:
            raise ValidationError(
                {self.type_field: ["Unsupported value: %s" % data_type]}
            )

        schema = type_schema if isinstance(type_schema, Schema) else type_schema()

        schema.context.update(context)

        if schema is self or self.__class__ is type_schema:
            # Pop the data type field and use super().load
            if self.type_field in data and self.type_field_remove:
                data.pop(self.type_field)
            return super().load(data, many=False, partial=partial, unknown=unknown)
        else:
            return schema.load(data, many=False, partial=partial, unknown=unknown)

    def validate(self, data, *, many=None, partial=None):
        try:
            self.load(data, many=many, partial=partial)
        except ValidationError as ve:
            return ve.messages
        return {}


""" Base models that allow persistency and indexing using the spintop.persistence modules.
"""
from datetime import datetime
from typing import List, Dict, Optional, Any, NamedTuple
from .base import BaseDataClass, model_property
from .serialization import SerializedWrapper, TYPE_KEY
from .queries import BaseQuery
from .meta import FieldsOf

from spintop.utils import utcnow_aware, is_aware

class PersistenceIDRecord(BaseDataClass):
    uuid: str
    is_deleted: bool = False
    tags: Dict[str, Optional[Any]] = dict
    start_datetime: datetime = utcnow_aware

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.start_datetime and not is_aware(self.start_datetime):
            raise ValueError('Please provide a tz-aware datetime.')


class PersistenceRecord(BaseDataClass):
    # Common to all
    uuid: str = model_property(fget=lambda record: record.index.uuid)

    # Same attributes, type should be changed by sub classes.
    index: PersistenceIDRecord = None
    data: BaseDataClass = None

class SerializedPersistenceRecord(SerializedWrapper):
    pass

index_fields = PersistenceIDRecord.fields()

class PersistenceRecordCollection(BaseDataClass):
    records: List[PersistenceRecord] = list

    def __init__(self, records=None):
        records = list(records) if records else []
        super().__init__(records=sorted(records, key=lambda record: index_fields.uuid.get_value(record.index)))
    
    @property
    def collector(self):
        return self.add_record

    def add_record(self, record):
        self.records.append(record)
        
    def apply(self, *fns):
        for record in self.iter_records():
            for fn in fns:
                fn(record)
                
    def iter_records(self):
        for record in self.records:
            yield record
                
    def count_unique(self, key=lambda record: record.uuid):
        occurances = set()
        self.apply(lambda record: occurances.add(key(record)))
        return len(occurances)
            
    def sort(self, key=lambda record: record.index.start_datetime):
        self.records.sort(key=key)
        
    def __eq__(self, other):
        return self.records == other.records

index_fields = PersistenceIDRecord.fields()

class Query(BaseQuery):
    
    def default_parts(self):
        yield index_fields.is_deleted == False


import os
from dataclasses import dataclass

from .api_client import SpintopAPIClientModule, SpintopAPISpecAuthBootstrap, create_backend_auth_bootstrap_factory
from .auth import FilePathCredentialsStore
from .spintop import SpintopFactory
from .api_client import SpintopAPIPersistenceFacade

from .storage import SITE_DATA_DIR
from .messages import SpintopMessagePublisher

def alias_for(env_name):
    return property(
        fget= lambda env: env[env_name],
        fset= lambda env, value: env.__setitem__(env_name, value)
    )

NO_VALUE = object()

def Spintop(api_url=None, verbose=False, init_env={}):
    env = SpintopEnv(init_env, verbose=verbose, api_url=api_url)
    return env.spintop_factory()

class SpintopEnv(object):
    # Credentials file is used to store access and refresh tokens.
    SPINTOP_CREDENTIALS_FILE: str = os.path.join(SITE_DATA_DIR, '.spintop-credentials.yml')

    # Client id and secret is used for machine to machine auth flow.
    SPINTOP_M2M_CLIENT_ID: str = None
    SPINTOP_M2M_SECRET_KEY: str = None

    SPINTOP_PERSISTENCE_TYPE: str = 'api'

    # used if SPINTOP_PERSISTENCE_TYPE is 'api'
    SPINTOP_API_URI: str = 'https://cloud.spintop.io'

    # used if SPINTOP_PERSISTENCE_TYPE is 'mongo'
    SPINTOP_MONGO_URI: str 

    # used if SPINTOP_PERSISTENCE_TYPE is 'postgres'
    SPINTOP_POSTGRES_URI: str

    # used by all persistence types
    SPINTOP_DATABASE_NAME: str = None

    # other
    SPINTOP_VERBOSE: bool = False

    # Aliases for python-friendly attributes
    api_url = alias_for('SPINTOP_API_URI') # URL
    api_uri = alias_for('SPINTOP_API_URI') # URI

    credentials_filepath = alias_for('SPINTOP_CREDENTIALS_FILE')
    selected_org_id = alias_for('SPINTOP_DATABASE_NAME')
    verbose = alias_for('SPINTOP_VERBOSE')
    database_name = alias_for('SPINTOP_DATABASE_NAME')

    def __init__(self, _init_values=None, verbose=False, api_url=None, ignore_invalid_init_value=False):
        if _init_values is None:
            _init_values = {}

        # Replace default values by possible a real env value.
        for key in self.ENV_NAMES:
            default_value = NO_VALUE
            if hasattr(self, key):
                default_value = self[key]

            setattr(self, key, os.environ.get(key, default_value))

        for key, value in _init_values.items():
            try:
                self[key] = value # Will validate if key is part of the support env variables.
            except KeyError:
                if not ignore_invalid_init_value:
                    raise

        if api_url:
            # for some reason, the requests module adds 1 second to every request made using localhost
            # replacing it with 127.0.0.1 removes this strange issue.
            self.api_url = api_url.replace('//localhost', '//127.0.0.1') 
        
        self.verbose = verbose

    @property
    def ENV_NAMES(self):
        return list(self.__annotations__.keys())

    def __getattr__(self, key):
        """When an attribute does not exist, attempt to retrieve it from env."""
        try:
            self._must_be_an_env_key(key)
            return os.environ[key]
        except KeyError as e:
            # Raise as attribute error for __getattr__ 
            raise AttributeError(str(e))

    def __getitem__(self, key):
        self._must_be_an_env_key(key)
        value = getattr(self, key)
        if value is NO_VALUE:
            raise KeyError(key)
        return value
    
    def __setitem__(self, key, value):
        self._must_be_an_env_key(key)
        setattr(self, key, value)

    def get(self, key, default_value=None):
        try:
            return self[key]
        except:
            return default_value

    def _must_be_an_env_key(self, key):
        if key not in self.ENV_NAMES:
            raise KeyError(f'{key!r} is not a SpintopEnv variable.')

    def freeze(self, specific_keys=None):
        if not specific_keys: specific_keys = self.ENV_NAMES
        return {key: self.get(key) for key in specific_keys}

    def freeze_database_access_only(self):
        keys = [
            'SPINTOP_PERSISTENCE_TYPE',
            self._facade_uri_env_name(self.SPINTOP_PERSISTENCE_TYPE),
            'SPINTOP_DATABASE_NAME'
        ]
        return self.freeze(keys)

    def copy(self, **new_env_values):
        new_env = SpintopEnv(self.freeze())
        for key, value in new_env_values.items():
            new_env[key] = value
        return new_env

    def new_database_context(self, database_name):
        env = self.copy()
        env.database_name = database_name
        return env

    def get_facade_uri(self, facade_type=None):
        return self[self._facade_uri_env_name(facade_type)]

    def transform_facade_uri(self, transform, facade_type=None):
        key_name = self._facade_uri_env_name(facade_type)
        self[key_name] = transform(self[key_name])

    def _facade_uri_env_name(self, facade_type=None):
        if facade_type is None: facade_type = self.SPINTOP_PERSISTENCE_TYPE
        return f'SPINTOP_{facade_type.upper()}_URI'

    def get_persistence_cls(self, facade_type=None):
        from .persistence.mongo import MongoPersistenceFacade
        from .persistence.postgresql import PostgreSQLPersistenceFacade
        if facade_type is None: facade_type = self.SPINTOP_PERSISTENCE_TYPE
        facade_cls_by_type = {
            'mongo': MongoPersistenceFacade,
            'postgres': PostgreSQLPersistenceFacade,
            'api': SpintopAPIPersistenceFacade,
        }

        if facade_type not in facade_cls_by_type:
            raise ValueError(f'Unknown facade type SPINTOP_PERSISTENCE_TYPE={facade_type!r}. Available: {list(facade_cls_by_type.keys())!r}')
        
        return facade_cls_by_type[facade_type]

    def get_persistence_params(self, facade_type=None):
        if facade_type is None: facade_type = self.SPINTOP_PERSISTENCE_TYPE
        return dict(
            uri=self.get_facade_uri(facade_type),
            database_name=self.database_name # shared by all
        )

    def persistence_facade_factory(self, message_publisher:SpintopMessagePublisher =None, facade_type=None):
        facade_cls = self.get_persistence_cls(facade_type)
        persistence_params = self.get_persistence_params(facade_type)
        facade = facade_cls.from_env(
            env=self,
            **persistence_params
        )
        if message_publisher is not None:
            facade.messages = message_publisher
        return facade
    
    def postgres_sql_ops_factory(self):
        from .persistence.postgresql import engine_from_uri, SQLOperations
        uri = self.get_facade_uri('postgres')
        engine = engine_from_uri(uri, database_name=self.database_name)
        return SQLOperations(engine)

    def postgres_singer_stream_factory(self, **kwargs):
        from .analytics.postgresql import PostgreSQLSingerTarget
        uri = self.get_facade_uri('postgres')
        return PostgreSQLSingerTarget(uri, database_name=self.database_name, **kwargs)

    def spintop_factory(self, **kwargs):
        auth_bootstrap_factory = SpintopAPISpecAuthBootstrap
        if self.SPINTOP_M2M_CLIENT_ID:
            # Use machine to machine auth bootstrap.
            auth_bootstrap_factory = create_backend_auth_bootstrap_factory(self.SPINTOP_M2M_CLIENT_ID, self.SPINTOP_M2M_SECRET_KEY)

        return SpintopFactory(
            api_url=self.api_uri,
            credentials_filepath=self.credentials_filepath,
            selected_org_id=self.selected_org_id,
            auth_bootstrap_factory=auth_bootstrap_factory,
            **kwargs
        )


from contextlib import contextmanager

from spintop.generators import Generator

from spintop.models import (
    PersistenceRecord,
    PersistenceIDRecord,
    SerializedPersistenceRecord,
    Query,
    FieldsOf
)

from spintop.utils import AnonymousGetRecursive


from ..logs import _logger
from ..errors import SpintopException, spintop_warn
from ..messages import NoopMessagePublisher, PersistenceRecordsUpdateMessage

logger = _logger('persistence')

class MissingMapper(SpintopException):
    def __init__(self, cls):
        super().__init__("Mapper for class {!r} is mandatory.".format(cls))

class NoMapperForObject(SpintopException):
    def __init__(self, obj, mappers):
        super(NoMapperForObject, self).__init__(
            'There are no known mapper able to interact with obj {!r} of class {}. Declared mappers are: {}'.format(
                obj,
                obj.__class__,
                [cls.__name__ for cls in mappers]
            )
        )
        
class DuplicateMapperClassName(SpintopException):
    def __init__(self, objcls):
        super(DuplicateMapperClassName, self).__init__(
            'The name of the class {} is duplicate. The class name linked to a mapper must be unique.'.format(
                objcls,
            )
        )

class RecordNotFound(SpintopException):
    def __init__(self, uuid):
        super().__init__(f'Record with UUID {uuid!r} not found.')

class ManyRecordFound(SpintopException):
    def __init__(self, uuid, count):
        super().__init__(f'{count} Records where found with UUID {uuid!r}, but only one was expected.')

index_fields = FieldsOf(PersistenceIDRecord)

class PersistenceFacade(object):
    logger = logger
    def __init__(self, serializer, messages=None):
        self.serializer = serializer

        if messages is None:
            # Will discard all messages
            messages = NoopMessagePublisher()

        self.messages = messages

    @classmethod
    def from_env(cls, uri, database_name, env=None):
        raise NotImplementedError()

    def _publish_records_update(self, records):
        message = PersistenceRecordsUpdateMessage.create(
            updated_ids=[record.index for record in records]
        )
        self.messages.records_update.publish(message)

    def _publish_update_all(self):
        spintop_warn('Current update all deletes and re-creates all tests instead of targeting deleted uuids. This is very inefficient.')
        self.messages.records_update.publish(PersistenceRecordsUpdateMessage.create(
            update_all=True
        ))

    def serialize_barrier(self, raw_records):
        raw_records = list(raw_records)
        schema = self.serializer.schema(many=True)

        if raw_records and isinstance(raw_records[0], PersistenceRecord):
            records = schema.dump(raw_records)
        else:
            records = raw_records

        if records and isinstance(records[0], dict):
            return [SerializedPersistenceRecord(record) for record in records]
        else:
            return records

    def create(self, records):
        serialized_records = self.serialize_barrier(records)
        ret_val = self._create(serialized_records)
        self._publish_records_update(serialized_records)
        return ret_val

    def _create(self, records):
        raise NotImplementedError()
        
    def retrieve(self, query=None, deserialize=True, limit_range=None, with_data=True):
        if query is None: query = Query()
        records = self._retrieve(query, limit_range=limit_range, with_data=with_data)
        if deserialize:
            yield from self.serializer.schema(many=True, partial=not with_data).load(r.as_dict() for r in records)
        else:
            yield from records

    def _retrieve(self, query, limit_range=None, with_data=True):
        """Should return a generator of dictionnaries of the records."""
        raise NotImplementedError()
        
    def retrieve_one(self, uuid, deserialize=True):
        records = self.retrieve(Query(index_fields.uuid == uuid), deserialize=deserialize)
        records = list(records)

        count = len(records)
        if count < 1:
            raise RecordNotFound(uuid)
        elif count > 1:
            raise ManyRecordFound(uuid, count)
        else:
            return records[0]
    
    def count(self, query=None):
        if query is None: query = Query()
        return self._count(query)
    
    def _count(self, query):
        raise NotImplementedError()

    def update(self, records, upsert=False):
        serialized_records = self.serialize_barrier(records)
        ret_val = self._update(serialized_records, upsert=upsert)
        self._publish_records_update(serialized_records)
        return ret_val

    def _update(self, records, upsert=False):
        raise NotImplementedError()
    
    def delete(self, query=None):
        if query is None: query = Query()

        # Needs list() to force retrieval before deletion.
        affected_records = list(self._retrieve(query, with_data=False))
        ret_val = self._delete(query)

        for record in affected_records:
            # We retrieved them before deletion, make sure to set the flag to True.
            index_fields.is_deleted.set_value(record.index, True)
            
        self._publish_records_update(affected_records)
        return ret_val

    def _delete(self, query=None):
        raise NotImplementedError()
    
    def create_records_generator(self):
        return PersistenceGenerator(self)

    @contextmanager
    def temporary_create(self, records):
        uuids = [record.uuid for record in records]
        try:
            self.create(records)
            yield uuids
        finally:
            self.delete(Query(index_fields.uuid <= uuids))

        
def create_mapper_name_index(mappers):
    mappers_name_index = {}
    for mapped_cls, mapper in mappers.items():
        name = mapped_cls.__name__
        if name in mappers_name_index:
            raise DuplicateMapperClassName(mapped_cls)
        mappers_name_index[name] = mapper
    return mappers_name_index
    
class PersistenceGenerator(Generator):
    def __init__(self, facade):
        super().__init__()
        self.facade = facade
    
    def __call__(self, *args, **kwargs):
        return self.facade.retrieve(*args, **kwargs)

class Mapper(object):
    def init(self):
        pass


from marshmallow import Schema, fields

class ManyRecordsSchema(Schema):
    records = fields.List(fields.Dict())
    
class RecordsCountSchema(Schema):
    count = fields.Int()

records_schema = ManyRecordsSchema()
record_count_schema = RecordsCountSchema()




import click
import webbrowser
from contextlib import contextmanager

from pprint import pprint

from ..errors import SpintopException, AuthUnauthorized, ExpiredAccessToken
from .schemas import credentials_schema

class AuthException(SpintopException):
    pass

REFRESH_MODULE = 'user'



class AuthModule(object):
    scopes = ['offline_access', 'openid', 'email', 'profile', 'authorization']
    def __init__(self, auth_bootstrap, credentials_store):
        self._auth_provider_accessor = auth_bootstrap.get_auth_provider if auth_bootstrap else None
        self.credentials_store = credentials_store
        self._reset_credentials()

    @property
    def credentials(self):
        self._init_check()
        return self._credentials
    
    @credentials.setter
    def credentials(self, value):
        self._credentials = value
    
    @property
    def user_orgs(self):
        self.assert_credentials()
        access_token = self.credentials.get('access_token', None)
        if access_token:
            return self.attempt_op_with_refresh(lambda: self.auth_provider.get_user_orgs(access_token))
        else:
            return []

    @property
    def auth_provider(self):
        if callable(self._auth_provider_accessor):
            return self._auth_provider_accessor()
        else:
            return self._auth_provider_accessor

    @auth_provider.setter
    def auth_provider(self, value):
        self._auth_provider_accessor = value

    def _reset_credentials(self):
        self._credentials = None
        self._initialized = False

    def _init_check(self):
        if not self._initialized and not self._credentials:
            self.credentials = self.credentials_store.retrieve()

            if self._credentials is None:
                try:
                    self.refresh_credentials()
                except AuthException:
                    pass
            # Other init stuff should go here
            self._initialized = True
    
    def assert_credentials(self):
        if not self.credentials:
            raise AuthException("No credentials available. Please login.")
    
    def assert_no_login(self):
        if self.credentials:
            raise AuthException(
                f"You already have credentials stored for {self.credentials['username']}. "
                "Spintop currently only support a single user per PC. "
                "Please logout before logging back in."
            )

    @contextmanager
    def login_device_flow(self, scopes=[]):
        self.assert_no_login()
        device_flow = self.auth_provider.DeviceAuthorizationFlow()
        response = device_flow.request_verification_url(scopes=self.scopes + scopes)
        yield response
        webbrowser.open(response.verification_uri_complete, new=2)
        content = device_flow.wait_for_approval()
        user_info = self.auth_provider.get_user_info(content['access_token'])
        self._save_access_token(content, username=user_info['email'])

    def get_user_info(self):
        self.assert_credentials()
        return self.auth_provider.get_user_info(self.credentials['access_token'])

    def login_user_pass(self, username, password, scopes=[]):
        self.assert_no_login()
        content = self.auth_provider.authenticate(username, password, scopes=self.scopes + scopes)
        self._save_access_token(content, username=username)

    def _save_access_token(self, content, username):
        self.credentials = credentials_schema.load({
            'username': username,
            'access_token': content.get('access_token'),
            'refresh_token': content.get('refresh_token'),
            'org_id': None,
            'refresh_module': REFRESH_MODULE
        })
        self.save_credentials()
        
    @property
    def credentials_key(self):
        
        key = self.credentials.get('username', None)
            
        if key is None:
            key = self.credentials.get('org_id', None)
            
        return key
        
    def save_credentials(self):
        self.credentials_store.store(self.credentials)
    
    def attempt_op_with_refresh(self, op):
        try:
            return op()
        except ExpiredAccessToken:
            self.refresh_credentials()
            return op()

    def refresh_credentials(self):
        try:
            self.credentials = self._refresh_credentials_obj(self._credentials)
        except AuthUnauthorized as e:
            raise AuthException(f'Unable to refresh credentials: {str(e)} Are you logged in?')
        self.save_credentials()
        
    def logout(self):
        # Add SpintopAPI-related logout stuff
        if self.credentials and self.credentials['refresh_token']:
            self.auth_provider.revoke_refresh_token(self.credentials['refresh_token'])
            
        self.credentials_store.delete()
        self._reset_credentials()
    
    def _refresh_credentials_obj(self, credentials):
        # username and refresh_token both stay valid
        if credentials is None:
            credentials = {}
        
        username = credentials.get('username', 'unknown')

        refresh_token = credentials.get('refresh_token', None)
        
        content = self.auth_provider.refresh_access_token(refresh_token)
        
        return credentials_schema.load({
            'username': username,
            'refresh_token': refresh_token,
            'access_token': content.get('access_token'),
            'org_id': None,
            'refresh_module': REFRESH_MODULE
        })
        
    def get_auth_headers(self):
        if self.credentials:
            return self.auth_provider.get_auth_headers(self.credentials.get('access_token'))
        else:
            return {}
        
    



import os
import uuid
from collections import OrderedDict

from spintop.persistence import PersistenceFacade
from spintop.models.persistence import (
    PersistenceIDRecord, 
    PersistenceRecord, 
    PersistenceRecordCollection
)

from spintop.models import (
    Query, 
    serialized_get_test_uuid,
    get_json_serializer,
    FieldsOf
)

from spintop.errors import SpintopException

from .models import RecordIndex, DataBlob
from .operations import engine_from_uri, SQLOperations
from .queryset import PostgreSQLQuery

index_fields = FieldsOf(PersistenceIDRecord)

class PostgreSQLPersistenceFacade(PersistenceFacade):
    def __init__(self, sql_engine, group_name='default'):
        self.sql_ops = SQLOperations(sql_engine)
        self.serializer = get_json_serializer()
        super().__init__(self.serializer)
        self.init()
        self.group = group_name

    def init(self):
        self.sql_ops.create_tables()
    
    @classmethod
    def from_env(cls, uri, database_name, env=None):
        # uri contains database, group_name defines the database tenant.
        return cls(engine_from_uri(uri), group_name=database_name)
        
    def _create(self, records):
        self._update(records)
    
    def add_records_to_session(self, session, records):
        # Create the data blobs
        data_blobs = [None]*len(records)
        for idx, record in enumerate(records):
            if not record.index['is_deleted']:
                data_blob = DataBlob(data=record.data, _group=self.group, blob_id=uuid.uuid4())
            data_blobs[idx] = data_blob

        session.bulk_save_objects(data_blobs)

        
        uuids_to_indices = {}
        
        for record, data_blob in zip(records, data_blobs):
            uuids_to_indices[record.uuid] = RecordIndex.create(
                record,
                _group=self.group,
                data_blob_id=data_blob.blob_id if data_blob else None
            )

        uuids = list(uuids_to_indices.keys())

        # Query these uuids, deleted or not.
        existing_query = Query(
            index_fields.uuid <= uuids, 
            index_fields.is_deleted.any_()
        )

        existing_indices = list(self._index_query_filter(session, existing_query).all())
        for record_index in existing_indices:
            # Only merge those records which already exist in the database
            session.merge(uuids_to_indices.pop(record_index.uuid))

        # Only add the records which did not exist in the database
        session.add_all(uuids_to_indices.values())

    def count(self, query=None):
        if query is None: query = Query()
        with self.sql_ops.new_session() as session:
            operation = session.query(RecordIndex).filter(*self._to_filter(query))
            return operation.count()
    
    def avg_obj_size(self):
        raise NotImplementedError()
        
    def _retrieve(self, query=None, limit_range=None, with_data=True):
        """ Retrieve is responsible with returning the test_record (top level phase) associated
        with the matched features.
        """
        with self.sql_ops.new_session() as session:
            if query is None: query = Query()
            
            operation = self._index_query_filter(session, query).order_by(RecordIndex.uuid.asc())
            
            if with_data:
                operation = operation.join(RecordIndex.data_blob)

            if limit_range:
                operation = operation.offset(limit_range[0]).limit(limit_range[1] - limit_range[0])

            yield from (record.to_serialized_record(with_data=with_data) for record in operation.all())

    def _index_query_filter(self, session, query):
        return session.query(RecordIndex).filter(*self._to_filter(query))

    def _to_filter(self, query):
        return PostgreSQLQuery(query).build_data_query(RecordIndex) + [RecordIndex._group == self.group]

    def _update(self, records, upsert=False):
        uuids = [record.uuid for record in records]

        with self.sql_ops.new_session() as session:
            # Delete referenced
            self._delete_query_from_session(session, Query(index_fields.uuid <= uuids))
            # Then update
            self.add_records_to_session(session, records)
            session.commit()
    
    def _delete_query_from_session(self, session, query):
        filters = self._to_filter(query)

        # Filter the datablobs by the ids referenced by indices of filters
        # Delete them.
        # The cascading effect will set the data_blob_id of indices to NULL.
        session.query(DataBlob).filter(DataBlob.blob_id == RecordIndex.data_blob_id, *filters).delete(synchronize_session=False)

        # Set the indices is_deleted = True
        session.query(RecordIndex).filter(*filters).update(
            {RecordIndex.is_deleted: True},
            synchronize_session=False
        )

    def _delete(self, match_query):
        with self.sql_ops.new_session() as session:
            self._delete_query_from_session(session, match_query)
            session.commit()



import re
from sqlalchemy import or_, and_

from spintop.models.persistence import PersistenceIDRecord
from spintop.utils import AnonymousGetRecursive

from .models import top_level_fields

top_level_fields_names = {field.name_ for field in top_level_fields}

class PostgreSQLQuery(object):
    def __init__(self, query):
        self.orig_query = query

    def build_data_query(self, record_table):
        queries = []

        for key, value in self.orig_query.value_equals.items():
            # data is stored as test_record: 
            if key in top_level_fields_names:
                queries.append(getattr(record_table, key) == value)
            else:
                field = self._get_index_field(record_table, key)
                if isinstance(value, re.Pattern):
                    queries.append(field.astext.op("~")(value.pattern))
                elif isinstance(value, bool):
                    # False -> false
                    queries.append(field.astext == str(value).lower())
                else:
                    queries.append(field.astext == str(value))
        
        for key, value in self.orig_query.list_contains.items():
            raise NotImplementedError()

        for key, list_of_values in self.orig_query.value_equals_one_of.items():
            if key in top_level_fields_names:
                queries.append(getattr(record_table, key).in_(list_of_values))
            else:
                field = self._get_index_field(record_table, key)
                queries.append(field.astext.in_(list_of_values))

        return queries

    def _get_index_field(self, record_table, field_name):
        field = AnonymousGetRecursive(record_table.index)
        return field.get_recursive(field_name)


from datetime import datetime

from spintop.persistence.base import PersistenceFacade

from ..models import PersistenceRecordCollection, Query, get_json_serializer, SerializedPersistenceRecord

from .schemas import records_schema, record_count_schema

class SpintopAPIPersistenceFacade(PersistenceFacade):
    def __init__(self, spintop_api):
        self.spintop_api = spintop_api
        self.serializer = get_json_serializer()
        super().__init__(self.serializer)

    @classmethod
    def from_env(self, uri, database_name=None, env=None):
        # database_name is the org_id
        api = env.spintop_factory()
        return api.records

    @property
    def session(self):
        return self.spintop_api.session

    def _create(self, records):
        serialized = self._records_in_schema(records)
        return self.session.post(self.spintop_api.get_link('records.create'), json=serialized)
        
    def _records_in_schema(self, records):
        return records_schema.dump({'records': [record.as_dict() for record in records]})

    def _retrieve(self, query, limit_range=None, with_data=True):
        query_dict = query.as_dict()
        if limit_range:
            query_dict['limit_range_inf'] = limit_range[0]
            query_dict['limit_range_sup'] = limit_range[1]
            
        query_dict['with_data'] = with_data

        resp = self.session.post(self.spintop_api.get_link('records.search'), json=query_dict)
        records = records_schema.load(resp.json())['records']
        return (SerializedPersistenceRecord(record) for record in records)
        
    def _count(self, query):
        query_dict = query.as_dict()
        resp = self.session.post(self.spintop_api.get_link('records.count'), json=query_dict)
        return record_count_schema.load(resp.json())['count']

    def _update(self, records, upsert=True):
        serialized = self._records_in_schema(records)
        return self.session.put(self.spintop_api.get_link('records.update'), json=serialized)
    
    def _delete(self, query):
        query_dict = query.as_dict()
        return self.session.delete(self.spintop_api.get_link('records.delete'), params=query_dict)
        
    def __str__(self):
        return f'{self.__class__.__name__}(url={self.spintop_api.api_url})'


""" Storage-friendly internal data classes. """
import numbers

from collections import defaultdict
from collections.abc import Mapping
from base64 import urlsafe_b64decode, urlsafe_b64encode
from uuid import UUID, uuid4

from datetime import datetime
from dataclasses import dataclass, fields, _MISSING_TYPE, MISSING, field, asdict
from typing import Union, List, Optional

from ..serialization import cls_of_obj

from ..base import (
    BaseDataClass, 
    _fields_cache,
    model_property
)

from ..persistence import PersistenceIDRecord

from ..view import ComplexPrimitiveView

from spintop.utils import isnan, utcnow_aware

NAME_SEPARATOR = ':'
NO_VERSION = '0'

def create_uuid():
    return uuid4()

def uuid_to_slug(uuidstring):
    try:
        uuid = UUID(uuidstring)
    except AttributeError:
        uuid = uuidstring
    
    return urlsafe_b64encode(uuid.bytes).rstrip(b'=').decode('ascii')

def test_uuid_generator(testbench_record=None, start_datetime=None, dut_record=None):
    datetime_format = None

    if start_datetime:
        datetime_format = start_datetime.strftime("%Y%m%d-%H%M%S-%f")

    if not datetime_format:
        # Not enough info for unique id. Generate one.
        value =  uuid_to_slug(create_uuid())
    else:
        value =  f"{str(testbench_record)}.{datetime_format}"

    if dut_record.id:
        value = f"{value}.{dut_record.id}-{dut_record.version}"
    
    return value
    

def slug_to_uuid(slug):
    return str(UUID(bytes=urlsafe_b64decode(slug + '==')))

def serialized_get_test_uuid(record):
    try:
        return record['test_id']['test_uuid']
    except KeyError:
        raise ValueError("test_id.test_uuid not in record! {!r}".format(record))

def compute_stats(features):
    name_tuple_to_count_map = defaultdict(int)

    for feature in features:
        # For feature ('x', 'y', 'z'), increment ('x',), ('x', 'y') and ('x', 'y', 'z')
        name_tuple = feature.name_tuple

        for i in range(1, len(name_tuple)):
            sub_name = name_tuple[:i]
            name_tuple_to_count_map[sub_name] += 1

    for feature in features:
        feature.feature_count = name_tuple_to_count_map[feature.name_tuple]


class VersionIDRecord(BaseDataClass):
    id: Optional[str]
    version: Optional[str]
    
    @classmethod
    def create(cls, id_or_dict, version=None):
        if isinstance(id_or_dict, Mapping):
            id = id_or_dict.get('id', None)
            version = id_or_dict.get('version', None)
        else:
            id = id_or_dict
            
        if version is None:
            version = NO_VERSION
            
        return cls(
            id=id,
            version=version
        )
        
    def match(self, other):
        id_match = self.id == other.id if self.id is not None else True
        version_match = self.version == other.version if self.version is not None else True
        return id_match and version_match

    def __str__(self):
        if self.version and self.version != NO_VERSION:
            return f'{self.id}-{self.version}'
        elif self.id:
            return str(self.id)
        else:
            return 'default'

class TestbenchIDRecord(VersionIDRecord):
    pass

class DutIDRecord(VersionIDRecord):
    pass

class OutcomeData(BaseDataClass):
    message: str = ''
    is_pass: bool = True
    is_skip: bool = False
    is_abort: bool = False
    is_ignore: bool = False
    
    def __bool__(self):
        return self.is_pass and not self.is_abort

    def impose_upon(self, other_outcome):
        if not self.is_ignore:
            # Pass propagates by falseness. False will propagate up, not True
            other_outcome.is_pass = other_outcome.is_pass and self.is_pass

            # Abort propagates by trueness. True will propagate up, not False
            other_outcome.is_abort = other_outcome.is_abort or self.is_abort

            # Abort propagates by trueness. True will propagate up, not False
            other_outcome.is_skip = other_outcome.is_skip or self.is_skip

class TestIDRecord(PersistenceIDRecord):
    testbench: TestbenchIDRecord = TestbenchIDRecord
    jig: TestbenchIDRecord = TestbenchIDRecord   
    dut: DutIDRecord = DutIDRecord
    outcome: OutcomeData = OutcomeData

    def __post_init__(self):
        if self.uuid is None:
            self.uuid = test_uuid_generator(self.testbench, self.start_datetime, self.dut)

    @classmethod
    def create(cls, testbench, dut, start_datetime, tags=None):
        dut_record = DutIDRecord.create(dut)
        testbench_record = TestbenchIDRecord.create(testbench)
        return cls(
            testbench=testbench_record,
            dut=dut_record,
            uuid=None, # will be generated in post init
            start_datetime=start_datetime,
            tags= tags if tags else {}
        )

    @property
    def testbench_name(self) -> Optional[str]:
        return self.testbench.id

    @testbench_name.setter
    def testbench_name(self, value):
        self.testbench.id = value

    def add_tag(self, key, value=True):
        self.tags[key] = tag_value_sanitizer(value)
        
    def remove_tag(self, key):
        del self.tags[key]

def tag_value_sanitizer(value):
    if isinstance(value, float) and isnan(value):
        # NaN to None.
        value = None

    return value


class FeatureRecordNoData(BaseDataClass):
    """ Feature ID is determined by its name and the test_id's test_uuid.
    """
    oid: Optional[str]
    test_id: TestIDRecord = None
    name: str
    description: str = ''
    version: int = 0
    depth: int = 0
    index: int = 0
    feature_count: int = 0 # Number of children
    ancestors: List[str]
    original: bool = True
    
    @classmethod
    def defaults(cls, **others):
        others['name'] = others.get('name', '')
        others['ancestors'] = others.get('ancestors', [])
        return super().defaults(**others)

    @classmethod
    def data_dataclass_fields(cls):
        """ The data fields are all fields except the ones declare up to this cls
        in the cls MRO.
        
        If a subclass defines only a dataclass field 'foo', this method will return
        foo only as a data field.
        """
        return set(_fields_cache.retrieve(cls)) - set(_fields_cache.retrieve(FeatureRecordNoData))
    
    @property
    def name_tuple(self):
        return tuple(self.ancestors) + (self.name,)

    @property
    def complete_name(self) -> str:
        return NAME_SEPARATOR.join(self.name_tuple)
    
class FeatureRecord(FeatureRecordNoData):
    user_data: dict = dict
    outcome: OutcomeData = OutcomeData
    
    @classmethod
    def defaults(cls, **others):
        others['user_data'] = others.get('user_data', {})
        others['feature_count'] = others.get('feature_count', 0)
        return super(FeatureRecord, cls).defaults(**others)
    
class PhaseFeatureRecord(FeatureRecord):
    duration: Optional[float] = None
    
class MeasureFeatureRecord(FeatureRecord):
    value_f: Optional[float]
    value_s: Optional[str]

    @property
    def value(self) -> Union[float, str, None]:
        if self.value_f is not None:
            return self.value_f
        else:
            return self.value_s

    @value.setter
    def value(self, value):
        if value is None or isinstance(value, str):
            self.value_f = None
            self.value_s = value
        elif isinstance(value, numbers.Number):
            self.value_f = value
            self.value_s = None
        else:
            raise ValueError('Only string or numeric types are supported. Received {!r} of type {}'.format(value, type(value)))
    
    
class TestRecordSummary(PhaseFeatureRecord):

    @property
    def total_feature_count(self):
        return self.feature_count

    @classmethod
    def data_dataclass_fields(cls):
        base = super(TestRecordSummary, cls).data_dataclass_fields()
        return list(base) + [FeatureRecord.fields().test_id.field_]
    
class DutOp(BaseDataClass):
    dut_match: DutIDRecord
    dut_after: DutIDRecord
    op_datetime: datetime = utcnow_aware
    
    @classmethod
    def create(cls, dut_match, dut_after, op_datetime):
        return cls(
            dut_match=DutIDRecord.create(dut_match),
            dut_after=DutIDRecord.create(dut_after),
            op_datetime=op_datetime
        )
    
    def does_applies_to(self, dut, on_datetime):
        return self.dut_match.match(dut) and self.op_datetime < on_datetime
    
    def apply_or_return(self, dut, on_datetime):
        if self.does_applies_to(dut, on_datetime):
            return self.apply(dut)
        else:
            return dut
        
    def apply(self, dut):
        """Apply op to dut. Does not check for datetime. """
        new_dut = self.dut_after.copy()
        if new_dut.id is None:
            new_dut.id = dut.id
        return new_dut
        
        
def default_cls_missing_view_fn(data_obj):
    cls = cls_of_obj(data_obj)
    return cls.get_default_view()
        
class DefaultPrimitiveView(ComplexPrimitiveView):
    def __init__(self):
        super(DefaultPrimitiveView, self).__init__(
            BaseDataClass, 
            cls_missing_view_fn=default_cls_missing_view_fn
        )



import socket
from contextlib import closing

from dateutil.tz import tzlocal
import pytz
from datetime import datetime

from collections.abc import Mapping
from contextlib import contextmanager
import cProfile, pstats, io

import yaml

try:
    import numpy as np
    isnan = np.isnan
except ImportError:
    isnan = lambda x: x is None

def utcnow_aware():
    dt = datetime.utcnow()
    return dt.replace(tzinfo=pytz.utc)

def is_aware(dt):
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None

def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]

def load_yaml_file(filename):
    with open(filename,'r') as readfile:
        data = readfile.read()
        return yaml.safe_load(data)
    
def write_yaml_file(filename, content):
    with open(filename,'w+') as writefile:
        yaml.dump(content, writefile)

def local_tz():
    return tzlocal()

@contextmanager
def profiling(output_filename):
    pr = cProfile.Profile()
    pr.enable()
    try:
        yield
    finally:
        pr.disable()
        s = io.StringIO()
        sortby = 'cumulative'
        with open(output_filename, 'w+') as profile_stream:
            ps = pstats.Stats(pr, stream=profile_stream).sort_stats(sortby)
            ps.print_stats()
            
class classproperty(object):
    def __init__(self, f):
        self.f = f
    def __get__(self, obj, owner):
        return self.f(owner)


import random
import time
import datetime

from spintop.models import SpintopTreeTestRecordBuilder
from spintop.utils import utcnow_aware
from .base import Generator

def normal(mu, sigma):
    return random.normalvariate(mu, sigma)

def random_failure(failure_odd):
    return random.uniform(0, 1.0) < failure_odd

base_mu_gen = lambda index: 5+(index*0.25)
base_sigma_gen = lambda index: 2
test_name_gen = lambda index: 'basicgen_test'
dut_id_gen = lambda index: 'dut%d' % ((index%25) if 17 < (index%25) < 20 else index % 17)
duration_gen = lambda index: 1 if (index%25) == 10 or (index%25) == 17 else 0.2


class RandomTestGenerator(Generator):
    def __call__(self, count=25, start_datetime=None, test_phases=('phase1', 'phase2'), measures=(), failure_rate=0.1):
        
        if start_datetime is None:
            start_datetime = utcnow_aware()
            
        for index in range(count):
            builder = SpintopTreeTestRecordBuilder()
            duration = duration_gen(index)
            fails = random_failure(failure_rate)
            builder.set_top_level_information(
                start_datetime=start_datetime,
                dut_id=dut_id_gen(index),
                testbench_name=test_name_gen(index),
                duration=duration,
                outcome=dict(
                    is_pass=not fails
                )
            )
            for test_name in test_phases:
                builder.new_phase(test_name, not fails, duration=duration/2)

            for measure_name in measures:
                builder.new_measure(measure_name, True, random.random())

            start_datetime += datetime.timedelta(seconds=duration)
            yield builder.build()


from spintop.analytics import AbstractSingerTarget

class SpintopAPIAnalytics(AbstractSingerTarget):
    def __init__(self, spintop_api):
        super().__init__()
        self.spintop_api = spintop_api

    @property
    def session(self):
        return self.spintop_api.session

    def send_messages(self, messages_str):
        return self.session.put(self.spintop_api.get_link('analytics.stream_update'), json=messages_str)



import os
import zipfile
from hashlib import md5

from glob import glob

from .base import Generator

class FileGenerator(Generator):
    def __call__(self, *filenames):
        for filename in filenames:
            with open(filename, 'rb') as subfile:
                yield subfile
    

class ZipFilesGenerator(Generator):
    
    def __call__(self, source):
        # We instantiate our OpenHTF test with the phases we want to run as args.
        zip_file = zipfile.ZipFile(source, "r")
        filenames = list(zip_file.namelist())
        for index, name in enumerate(filenames):
            print('Processing file %s. %d/%d' % (name, index, len(filenames)))
            with zip_file.open(name) as subfile:
                yield subfile

    def hash_fn(self, source):
        m = md5()
        with open(source, "rb") as f:
            data = f.read() #read file in chunk and call update on each chunk if file is large.
            m.update(data)
        return m.hexdigest()


import dateutil.parser
import pytz

from copy import copy
from typing import Union, NamedTuple
from collections import OrderedDict
from collections.abc import Sequence, Mapping, MappingView
from functools import lru_cache
from datetime import datetime

from marshmallow import Schema, fields, pre_load, post_dump, INCLUDE

from spintop.utils.marshmallow_dataclass import class_schema
from spintop.utils.marshmallow_oneofschema import OneOfSchema

from spintop.utils import local_tz, isnan, dict_ops


SCHEMA_TYPES = {'json', 'bson'}

_TYPE_TO_CLS_MAP = {}

_TYPE_TO_SCHEMA_BY_TYPE = {schema_type: {} for schema_type in SCHEMA_TYPES}

TYPE_KEY = '_type'

def register_type(cls, _type=None, schema_cls=None):
    if _type is None:
        _type = cls.__name__
    cls.__type = _type
    
    if _type in _TYPE_TO_CLS_MAP and _TYPE_TO_CLS_MAP[_type] != cls:
        raise ValueError("Type %s is already registered to class %s" % (_type, _TYPE_TO_CLS_MAP[_type]))

    _TYPE_TO_CLS_MAP[_type] = cls
    for schema_type in _TYPE_TO_SCHEMA_BY_TYPE:
        schema_cls = DataClassSchema(cls, schema_type=schema_type, schema_cls=schema_cls)

        # Building the schema now improves serialization performance by up to 800%
        _TYPE_TO_SCHEMA_BY_TYPE[schema_type][_type] = schema_cls()

def cls_of_obj(obj):
    if isinstance(obj, Mapping):
        return cls_of_serialized(obj)
    else:
        return obj.__class__
    # try:
    #     _type = type_of(obj)
    #     cls = cls_of_type(_type)
    #     if cls is None:
    #         raise AttributeError(f'{cls} is not mapped to a type')
    #     return cls
    # except AttributeError:
    #     return cls_of_serialized(obj)

def cls_of_serialized(obj_serialized):
    return cls_of_type(serialized_type_of(obj_serialized))

def cls_of_type(_type):
    return _TYPE_TO_CLS_MAP.get(_type, None)

def type_of(cls):
    return cls.__type

def serialized_type_of(obj_serialized):
    return obj_serialized[TYPE_KEY]

def type_dict_of(cls):
    return {TYPE_KEY: type_of(cls)}

def is_type_of(cls, base_cls):
    return type_of(cls) == type_of(base_cls)

def is_serialized_type_of(obj_serialized, base_cls):
    return serialized_type_of(obj_serialized) == type_of(base_cls)

def get_serializer():
    return get_bson_serializer()

def get_bson_serializer():
    return Serializer(get_base_schema_cls('bson'))

def get_json_serializer():
    return Serializer(get_base_schema_cls('json'))

@lru_cache(maxsize=None)
def DataClassSchema(cls, schema_type=None, schema_cls=None):
    base_schema_cls = get_base_schema_cls(schema_type)
    if schema_cls:
        schema_cls = compose_schema_cls(base_schema_cls, schema_cls)
    else:
        schema_cls = base_schema_cls
    schema = class_schema(cls, base_schema=schema_cls)
    return schema

def get_base_schema_cls(schema_type):
    schema_map = {
        'json': JsonCompatSchema,
        'bson': BsonCompatSchema,
        None: JsonCompatSchema # default
    }
    try:
        return schema_map[schema_type]
    except KeyError:
        raise ValueError(f'Schema type {schema_type!r} not in supported list: {schema_map.keys()!r}')

def compose_schema_cls(base, child):
    return type(f'Composed{child.__name__}', (base, child), {})

class Serializer(object):
    def __init__(self, base_schema_cls):
        self.base_schema_cls = base_schema_cls

    def schema(self, *args, **kwargs):
        return self.base_schema_cls(*args, **kwargs)

    def serialize(self, obj):
        master_schema = self.schema()
        try:
            type_of(obj) # Tests if it has a __type attribute.
        except (AttributeError, TypeError):
            return self.serialize_primitive(obj)
        else:
            return master_schema.dump(obj)

    def deserialize(self, obj, cls=None):
        master_schema = self.schema()
        # _type = serialized_type_of(obj)
        if cls:
            # If no type, use cls as default.
            obj.update(type_dict_of(cls))
        return master_schema.load(obj)

    def serialize_primitive(self, obj):
        if isinstance(obj, Sequence) and not isinstance(obj, str):
            return [self.serialize(subobj) for subobj in obj]
        elif isinstance(obj, Mapping):
            return {key: self.serialize(value) for key, value in obj.items()}
        else:
            try:
                Field = self.base_schema_cls.TYPE_MAPPING[type(obj)]
            except KeyError:
                # No mapped field. Return as-is !
                return obj
            field = Field()
            return field._serialize(obj, None, None)

    def deserialize_primitive(self, obj, cls=None):
        if isinstance(obj, Sequence):
            return [self.deserialize(subobj, cls=cls) for subobj in obj]
        elif isinstance(obj, Mapping):
            return {key: self.deserialize(value, cls=cls) for key, value in obj.items()}
        else:
            raise ValueError('Unable to deserialize primitives other than list or dict.')


class NoopField(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        return value

class OrderedDictField(fields.Mapping):
    def _serialize(self, value, attr, obj, **kwargs):
        value_as_dict = dict(value)
        serialized_as_dict = super()._serialize(value_as_dict, attr, obj, **kwargs)
        return [(key, serialized_as_dict[key]) for key in value]

    def _deserialize(self, list_value, attr, data, **kwargs):
        value_as_dict = {key: value for key, value in list_value}
        deserialized_as_dict = super()._deserialize(value_as_dict, attr, data, **kwargs)
        result = OrderedDict()
        for key, _ in list_value:
            result[key] = deserialized_as_dict[key]
        return result

class FloatFieldNoNaN(fields.Float):
    def _serialize(self, value, attr, data, **kwargs):
        if value is None or isnan(value):
            value = None
        num = super()._serialize(value, attr, data, **kwargs)
        return num

class AwareDateTimeDeserializeNoop(fields.AwareDateTime):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, default_timezone=pytz.utc, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, datetime):
            return value.replace(tzinfo=self.default_timezone)
        else:
            return super()._deserialize(value, attr, data, **kwargs)

def base_type_mapping(additions={}):
    default_type_mapping = copy(Schema.TYPE_MAPPING)
    default_type_mapping.update({
        float: FloatFieldNoNaN,
        OrderedDict: OrderedDictField,
        datetime: AwareDateTimeDeserializeNoop
    })

    try:
        from pandas import Timestamp
    except ImportError:
        pass
    else:
        default_type_mapping.update({
            Timestamp: AwareDateTimeDeserializeNoop
        })

    default_type_mapping.update(additions)
    return default_type_mapping


class AbstractSchema(OneOfSchema):
    type_field = TYPE_KEY

    def get_obj_type(self, obj):
        # If a mapping, returns the class from the _type attribute
        # If an object, returns the class itself.
        cls = cls_of_obj(obj) 

        if cls is None:
            return None
        else:
            # Return the string registered to that cls.
            return type_of(cls) 

class JsonCompatSchema(AbstractSchema):
    type_schemas = _TYPE_TO_SCHEMA_BY_TYPE['json']
    TYPE_MAPPING = base_type_mapping()

class BsonCompatSchema(AbstractSchema):
    type_schemas = _TYPE_TO_SCHEMA_BY_TYPE['bson']
    # Bson supports native datetime objects.
    TYPE_MAPPING = base_type_mapping({
        datetime: NoopField # keeps datetime as is
    })

class SerializedWrapper(dict_ops.DictWrapper):
    
    @property
    def type_(self):
        return self[TYPE_KEY]

    @type_.setter
    def type_(self, value):
        self[TYPE_KEY] = value

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)



from collections.abc import Mapping
from collections import OrderedDict

from .meta import MetaField

from spintop.utils.dict_ops import (
    update, 
    deepen_value_in_dict,
    deepen_dict,
    flatten_dict,
    DEFAULT_PLACEHOLDER
)
from spintop.utils.get_recursive import GetRecursive

class _MetaFieldAccessor(object):
    def __init__(self, meta_field):
        self.field = meta_field

    def __call__(self, obj):
        return self.field.get_value(obj)

    def __repr__(self):
        return repr(self.field)

class _ValueAccessor(object):
    def __init__(self, value):
        self.value = value
    
    def __call__(self, obj):
        return self.value

    def __repr__(self):
        return f'Value({repr(self.value)})'
    
class _CallableAccessor(object):
    def __init__(self, fn):
        self.fn = fn

    def __call__(self, obj):
        return self.fn(GetRecursive(obj))

    def __repr__(self):
        return f'Call({self.fn.__name__}: {self.fn!r})'
    
class DataClassPrimitiveView(object):
    def __init__(self, key_values_mapping={}):
        self.build_accessors(key_values_mapping)
        
    def build_accessors(self, key_values_mapping):
        """ Creates a dict of callable
        """

        self._accessors_mapping = _build_accessors(key_values_mapping)
        
    def apply(self, data_obj, flatten_dict=False, attribute_error_policy='ignore'):
        result = {}

        raise_attribute_errors = attribute_error_policy != 'ignore'
        for accessor_key, accessor_value in self._accessors_mapping.items():
            # Set to none if does not exist.
            key = None

            try:
                key = _apply_single_accessor(accessor_key, data_obj)
                value = _apply_single_accessor(accessor_value, data_obj)
            except AttributeError as e:
                if raise_attribute_errors:
                    raise
                else:
                    _apply_key_value(result, key, DEFAULT_PLACEHOLDER, flatten_dict)
                    continue
            
            _apply_key_value(result, key, value, flatten_dict)
        
        return result

def _build_accessors(key_values_mapping):
    flat_key_values_mapping = flatten_dict(key_values_mapping)
    return {_build_single_accessor(key): _build_single_accessor(value) for key, value in flat_key_values_mapping.items()}

def _build_single_accessor(obj):
    if isinstance(obj, tuple):
        return tuple(_build_single_accessor(subobj) for subobj in obj)
    elif isinstance(obj, MetaField) or isinstance(obj, property):
        return _MetaFieldAccessor(obj)
    # elif isinstance(obj, property):
    #     return _PythonAccessor(obj)
    elif callable(obj):
        return _CallableAccessor(obj) # Obj is an accessor
    else:
        return _ValueAccessor(obj) # Will always return obj
    
def _apply_single_accessor(accessor, obj):
    if isinstance(accessor, tuple):
        return tuple(_apply_single_accessor(subaccessor, obj) for subaccessor in accessor)
    else:
        return accessor(obj)

def default_cls_missing_view_fn(data_obj):
    raise NotImplementedError()

def _apply_key_value(_dict, key, value, flatten_dict):
    if isinstance(key, tuple) and not flatten_dict:
        # Push value deep into a nested dictionnary
        # key = (x, y, z)
        # result.update({x: {y: {z: value}}})
        update(_dict, deepen_value_in_dict(key, value))
    else:
        # or keep as is (flat)
        # key = (x, y, z)
        # result[(x, y, z)] = value
        _dict[key] = value

class ComplexPrimitiveView(object):
    def __init__(self, base_cls, cls_missing_view_fn=None):
        if cls_missing_view_fn is None: cls_missing_view_fn = default_cls_missing_view_fn
        
        self.base_cls = base_cls
        self.cls_missing_view_fn = cls_missing_view_fn
        self.custom_views = {}
        
    def _find_closest_parent(self, cls):
        closest_parent = None
        
        if cls in self.custom_views:
            return cls
        
        for match_cls in self.custom_views:
            possible_parent = None
            if issubclass(cls, match_cls):
                possible_parent = match_cls
            
            if possible_parent and (closest_parent is None or issubclass(possible_parent, closest_parent)):
                # possible parent is closer to cls than the previous closest parent.
                closest_parent = possible_parent
                
        return closest_parent
    
    def _find_view(self, data_obj):
        cls = data_obj.__class__
        best_match = self._find_closest_parent(cls)
        if best_match:
            return self.custom_views[best_match]
        else:
            return self.cls_missing_view_fn(data_obj)
    
    def add_view(self, cls, view_mapping):
        self.custom_views[cls] = DataClassPrimitiveView(view_mapping)
        
    def apply(self, root_data_obj, flatten_dict=False, key_prefix=None, **apply_kwargs):
        flatten_key_prefix = None
        if flatten_dict:
            if key_prefix is None:
                flatten_key_prefix = ()
            else:
                flatten_key_prefix = key_prefix
        
        result = self._apply_data_obj(root_data_obj, flatten_key_prefix=flatten_key_prefix, **apply_kwargs)

        if not flatten_dict and key_prefix:
            return deepen_value_in_dict(key_prefix, result)
        else:
            return result
    
        
    
    def _apply_data_obj(self, data_obj, flatten_key_prefix=None, **apply_kwargs):
        mapping = self._find_view(data_obj).apply(data_obj, flatten_dict=flatten_key_prefix is not None, **apply_kwargs)
        
        flatten_key_prefix_built = _build_single_accessor(flatten_key_prefix)
        flatten_key_prefix_applied = _apply_single_accessor(flatten_key_prefix_built, data_obj)

        return self._apply_mapping(mapping, flatten_key_prefix=flatten_key_prefix_applied, **apply_kwargs)
    
    def _apply_mapping(self, mapping, flatten_key_prefix=None, **apply_kwargs):
        result = OrderedDict()
        
        for key, value in mapping.items():
            
            flat_key = None
            if flatten_key_prefix is not None:
                flat_key = join_key_prefix(flatten_key_prefix, key)
            
            discard_false_subvalue = False 
            
            if isinstance(value, self.base_cls):
                subvalue = self._apply_data_obj(value, flatten_key_prefix=flat_key, **apply_kwargs)
                # If bool(subvalue) is False, then the sub-view returned an empty mapping or None. Do not consider it.
                discard_false_subvalue = True 
            elif isinstance(value, Mapping):
                subvalue = self._apply_mapping(value, flatten_key_prefix=flat_key, **apply_kwargs)
                # If bool(subvalue) is False, the mapping is empty or none. Do not consider it.
                discard_false_subvalue = True 
            elif flat_key is not None:
                # limit case.
                # If flatten, we will .update the result instead of applying
                subvalue = {flat_key: value}
            else:
                subvalue = value
            
            if discard_false_subvalue and not subvalue:
                continue # discard
            
            if flat_key:
                for subkey in subvalue:
                    if subkey in result:
                        print("{} already in result.".format(subkey))
                update(result, subvalue)
            else:
                if key in result:
                    print("{} already in result.".format(key))
                update(result, {key: subvalue})
        
        return result

def join_key_prefix(key_prefix, key):
    if not isinstance(key, tuple):
        key = (key,)

    if not isinstance(key_prefix, tuple):
        key_prefix = (key_prefix,)
        
    return key_prefix + key




import os

from spintop.persistence import PersistenceFacade
from spintop.models import (
    SpintopTestRecord, 
    SpintopSerializedTestRecord, 
    SpintopSerializedTestRecordCollection, 
    TestRecordSummary,
    Query, 
    serialized_get_test_uuid
)
from spintop.errors import SpintopException

from .operations import db_from_mongo_uri
from .mappers import create_mappers, simple_mongo_mapper_factory, TestRecordSummary, FeatureRecord

class MongoPersistenceFacade(PersistenceFacade):
    def __init__(self, mongo_db, serializer=None):
        self.mongo_db = mongo_db
        self.serializer = serializer
        mappers = create_mappers(mongo_db, serializer=serializer)
        self.feature_mapper = mappers[FeatureRecord]
        self._init(mappers)
        self.dc_serializer = self.feature_mapper.dc_serializer

        super().__init__(self.dc_serializer.serializer)
    
    def _init(self, mappers):
        for _, mapper in mappers.items():
            mapper.init()

    @classmethod
    def from_env(cls, uri, database_name, env=None):
        return cls(db_from_mongo_uri(uri, database_name))
        
    def _create(self, records):
        self._iter_op('create', records)
        
    def _iter_op(self, op_name, records):
        all_features = []
        for record in records:
            all_features += record.all_features

        self.logger.info('Processing {} of {} features.'.format(
                op_name, len(all_features)
            )
        )
        getattr(self.feature_mapper, op_name)(all_features)
    
    def count(self, query=None):
        if query is None:
            query = Query()
        return self.feature_mapper.count(query)
    
    def avg_obj_size(self):
        # Count is based on test records, multiple avg feature size with avg feature count
        avg_feature_count = self.feature_mapper.count_total() / self.feature_mapper.count()
        return self.feature_mapper.collection_stats().get('avgObjSize') * avg_feature_count
        
    def _retrieve(self, query=None, limit_range=None):
        """ Retrieve is responsible with returning the test_record (top level phase) associated
        with the matched features.
        """
        query = self._include_top_level_in_query(query)
        all_features = self.feature_mapper.retrieve(query, limit_range=limit_range)

        return SpintopSerializedTestRecordCollection.from_features(all_features)

    def _include_top_level_in_query(self, query):
        if query is None:
            query = Query()
        return [query.create_top_level_query(), query]

    def _update(self, records, upsert=False):
        all_features = []
        test_uuids = []

        for record in records:
            all_features += record.all_features
            test_uuids += [record.test_uuid]

        self.feature_mapper.delete(Query().test_uuid_any_of(test_uuids))
        self.feature_mapper.create(all_features)
    
    def _delete(self, match_query):
        match_query = self._include_top_level_in_query(match_query)
        self.logger.info('Deleting flat records based on query: %s.' % repr(match_query))
        self.feature_mapper.delete(match_query)

    def create_mapper(self, name, **kwargs):
        return simple_mongo_mapper_factory(self.mongo_db, name, serializer=self.serializer, **kwargs)


from .test_records import *

from .base import (
    BaseDataClass,
    model_property,
    model_constant
)

from .meta import FieldsOf

from .persistence import (
    PersistenceIDRecord,
    PersistenceRecord,
    PersistenceRecordCollection,
    SerializedPersistenceRecord,
    Query
)

from .queries import multi_query_deserialize, multi_query_serialize

from .serialization import (
    get_serializer, 
    get_bson_serializer, 
    get_json_serializer,
    type_of,
    serialized_type_of,
    is_type_of,
    type_dict_of,
    is_serialized_type_of
)

from .view import DataClassPrimitiveView, ComplexPrimitiveView, update

from .view_helpers import (
    value_of_feature_named
)



import psycopg2
from urllib.parse import urlparse, parse_qs

from target_postgres.postgres import PostgresTarget
from target_postgres import target_tools

from .base import AbstractSingerTarget

class PostgreSQLSingerTarget(AbstractSingerTarget):
    def __init__(self, uri, database_name, config={}):
        super().__init__()
        result = urlparse(uri)
 
        username = result.username
        password = result.password
        hostname = result.hostname
        port = result.port

        # Support host being passed as query param. This is used by sqlalchemy
        # to allow unix sockets:
        # postgres://user:password@/dbname?host=/path/to/db
        query = parse_qs(result.query)

        if 'host' in query and not hostname:
            # host in query is a list of hosts.
            # join with commas
            hostname = ','.join(query['host'])

        self.connection = psycopg2.connect(
            database = database_name,
            user = username,
            password = password,
            host = hostname,
            port = port
        )
    
        self.config = config

    def send_messages(self, messages_str):
        with self.connection:
            postgres_target = PostgresTarget(
                self.connection,
                postgres_schema=self.config.get('postgres_schema', 'public'),
                logging_level=self.config.get('logging_level'),
                persist_empty_tables=self.config.get('persist_empty_tables'),
                add_upsert_indexes=self.config.get('add_upsert_indexes', True),
                before_run_sql=self.config.get('before_run_sql'),
                after_run_sql=self.config.get('after_run_sql'),
            )
            target_tools.stream_to_target(messages_str, postgres_target, config=self.config)